
/*
* Galois field arithmetic components for security application
* dual-rail/1-of-4 relaxed balancing
* using generic Verilog primitives
* !!! Estimation of physical parameters is based on AMS C35 0.35um2 library (should be changed) !!!
* @scaleenergy 1
* @scalearea 0.001
*
* Author: Ashur Rafiev
* July 2009
*/

/*
* GF(2) addition
* @energy 0.36
* @area 330
* @delay 2
* @switch 1
*/
module GF2ADD(Q, A, B);
	input [1:0] A, B;
	output [1:0] Q;

	wire w0, w1, w2, w3;
	nand g00(w0, A[0], B[0]);
	nand g01(w1, A[0], B[1]);
	nand g02(w2, A[1], B[0]);
	nand g03(w3, A[1], B[1]);

	nand g10(Q[0], w0, w3);
	nand g11(Q[1], w1, w2);
endmodule

/*
* GF(2)/GF(3) mixed-radix addition
* @energy 0.36
* @area 495
* @delay 2
* @switch 1
*/
module GF23ADD(Q, A, B);
	input [2:0] A;
	input [1:0] B;
	output [2:0] Q;

	wire w0, w1, w2, w3, w4, w5;
	nand g00(w0, A[0], B[0]);
	nand g01(w1, A[0], B[1]);
	nand g02(w2, A[1], B[0]);
	nand g03(w3, A[1], B[1]);
	nand g04(w4, A[2], B[0]);
	nand g05(w5, A[2], B[1]);

	nand g10(Q[0], w0, w5);
	nand g11(Q[1], w2, w1);
	nand g12(Q[2], w4, w3);
endmodule

/*
* GF(2)/GF(4) mixed-radix addition
* @energy 0.36
* @area 660
* @delay 2
* @switch 1
*/
module GF24ADD(Q, A, B);
	input [3:0] A;
	input [1:0] B;
	output [3:0] Q;

	wire w0, w1, w2, w3, w4, w5, w6, w7;
	nand g00(w0, A[0], B[0]);
	nand g01(w1, A[0], B[1]);
	nand g02(w2, A[1], B[0]);
	nand g03(w3, A[1], B[1]);
	nand g04(w4, A[2], B[0]);
	nand g05(w5, A[2], B[1]);
	nand g06(w6, A[3], B[0]);
	nand g07(w7, A[3], B[1]);

	nand g10(Q[0], w0, w3);
	nand g11(Q[1], w1, w2);
	nand g12(Q[2], w4, w7);
	nand g13(Q[3], w5, w6);
endmodule

/*
* GF(3) addition
* @energy 0.39
* @area 714
* @delay 2
* @switch 1
*/
module GF3ADD(Q, A, B);
	input [2:0] A, B;
	output [2:0] Q;

	wire w0, w1, w2, w3, w4, w5, w6, w7, w8;
	nand g00(w0, A[0], B[0]);
	nand g01(w1, A[1], B[2]);
	nand g02(w2, A[2], B[1]);
	nand g03(w3, A[0], B[1]);
	nand g04(w4, A[1], B[0]);
	nand g05(w5, A[2], B[2]);
	nand g06(w6, A[0], B[2]);
	nand g07(w7, A[1], B[1]);
	nand g08(w8, A[2], B[0]);

	nand g10(Q[0], w0, w1, w2);
	nand g11(Q[1], w3, w4, w5);
	nand g12(Q[2], w6, w7, w8);
endmodule

/*
* GF(4) addition
* @energy 0.42
* @area 1244
* @delay 2
* @switch 1
*/
module GF4ADD(Q, A, B);
	input [3:0] A, B;
	output [3:0] Q;

	wire w0, w1, w2, w3, w4, w5, w6, w7,
		w8, w9, w10, w11, w12, w13, w14, w15;
	nand g00(w0, A[0], B[0]);
	nand g01(w1, A[0], B[1]);
	nand g02(w2, A[0], B[2]);
	nand g03(w3, A[0], B[3]);
	nand g04(w4, A[1], B[0]);
	nand g05(w5, A[1], B[1]);
	nand g06(w6, A[1], B[2]);
	nand g07(w7, A[1], B[3]);
	nand g08(w8, A[2], B[0]);
	nand g09(w9, A[2], B[1]);
	nand g010(w10, A[2], B[2]);
	nand g011(w11, A[2], B[3]);
	nand g012(w12, A[3], B[0]);
	nand g013(w13, A[3], B[1]);
	nand g014(w14, A[3], B[2]);
	nand g015(w15, A[3], B[3]);

	nand g10(Q[0], w0, w5, w10, w15);
	nand g11(Q[1], w1, w4, w11, w14);
	nand g12(Q[2], w2, w7, w8, w13);
	nand g13(Q[3], w3, w6, w9, w12);
endmodule

/*
* Binary inversion
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF2ADD1(Q, A);
	input [1:0] A;
	output [1:0] Q;

	assign Q[0] = A[1];
	assign Q[1] = A[0];
endmodule

/*
* GF(3) inversion (x+1)
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF3ADD1(Q, A);
	input [2:0] A;
	output [2:0] Q;

	assign Q[0] = A[2];
	assign Q[1] = A[0];
	assign Q[2] = A[1];
endmodule

/*
* GF(3) inversion (x+2)
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF3ADD2(Q, A);
	input [2:0] A;
	output [2:0] Q;

	assign Q[0] = A[1];
	assign Q[1] = A[2];
	assign Q[2] = A[0];
endmodule

/*
* GF(4) inversion (x+1)
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF4ADD1(Q, A);
	input [3:0] A;
	output [3:0] Q;

	assign Q[0] = A[1];
	assign Q[1] = A[0];
	assign Q[2] = A[3];
	assign Q[3] = A[2];	
endmodule

/*
* GF(4) inversion (x+A)
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF4ADDA(Q, A);
	input [3:0] A;
	output [3:0] Q;

	assign Q[0] = A[2];
	assign Q[1] = A[3];
	assign Q[2] = A[0];
	assign Q[3] = A[1];
endmodule

/*
* GF(4) inversion (x+B)
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF4ADDB(Q, A);
	input [3:0] A;
	output [3:0] Q;

	assign Q[0] = A[3];
	assign Q[1] = A[2];
	assign Q[2] = A[1];
	assign Q[3] = A[0];
endmodule

/*
* GF(2) multiplication
* @energy 0.39
* @area 182
* @delay 2
* @switch 1
*/
module GF2MUL(Q, A, B);
	input [1:0] A, B;
	output [1:0] Q;

	wire w0, w1;
	nor g00(w0, A[0], B[0]);
	nand g01(w1, A[1], B[1]);

	not g10(Q[0], w0);
	not g11(Q[1], w1);
endmodule

/*
* GF(2)/GF(3) mixed-radix multiplication
* @energy 0.39
* @area 273
* @delay 2
* @switch 1
*/
module GF23MUL(Q, A, B);
	input [2:0] A;
	input [1:0] B;
	output [2:0] Q;

	wire w0, w1, w2;
	nor g00(w0, A[0], B[0]);
	nand g01(w1, A[1], B[1]);
	nand g02(w2, A[2], B[1]);

	not g10(Q[0], w0);
	not g11(Q[1], w1);
	not g12(Q[2], w2);
endmodule

/*
* GF(2)/GF(4) mixed-radix multiplication
* @energy 0.39
* @area 364
* @delay 2
* @switch 1
*/
module GF24MUL(Q, A, B);
	input [3:0] A;
	input [1:0] B;
	output [3:0] Q;

	wire w0, w1, w2, w3;
	nor g00(w0, A[0], B[0]);
	nand g01(w1, A[1], B[1]);
	nand g02(w2, A[2], B[1]);
	nand g03(w3, A[3], B[1]);

	not g10(Q[0], w0);
	not g11(Q[1], w1);
	not g12(Q[2], w2);
	not g13(Q[3], w3);
endmodule

/*
* GF(3) multiplication
* @energy 0.39
* @area 421
* @delay 2
* @switch 1
*/
module GF3MUL(Q, A, B);
	input [2:0] A, B;
	output [2:0] Q;
	
	wire wa0;
	nor ga0(wa0, A[0], B[0]);
	not ga1(Q[0], wa0);
	
	wire w0, w1, w2, w3;
	nand g00(w0, A[1], B[1]);
	nand g01(w1, A[2], B[2]);
	nand g02(w2, A[1], B[2]);
	nand g03(w3, A[2], B[1]);
	
	nand g10(Q[1], w0, w1);
	nand g11(Q[2], w2, w3);
endmodule

/*
* GF(4) multiplication
* @energy 0.39
* @area 805
* @delay 2
* @switch 1
*/
module GF4MUL(Q, A, B);
	input [3:0] A, B;
	output [3:0] Q;
	
	wire wa0;
	nor ga0(wa0, A[0], B[0]);
	not ga1(Q[0], wa0);
	
	wire w0, w1, w2, w3, w4, w5, w6, w7, w8;
	nand g00(w0, A[1], B[1]);
	nand g01(w1, A[2], B[3]);
	nand g02(w2, A[3], B[2]);
	nand g03(w3, A[1], B[2]);
	nand g04(w4, A[2], B[1]);
	nand g05(w5, A[3], B[3]);
	nand g06(w6, A[1], B[3]);
	nand g07(w7, A[2], B[2]);
	nand g08(w8, A[3], B[1]);
	
	nand g10(Q[1], w0, w1, w2);
	nand g11(Q[2], w3, w4, w5);
	nand g12(Q[3], w6, w7, w8);	
endmodule

/*
* GF(3) multiplication with constant 2
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF3MUL2(Q, A);
	input [2:0] A;
	output [2:0] Q;

	assign Q[0] = A[0];
	assign Q[1] = A[1];
	assign Q[2] = A[2];
endmodule

/*
* GF(4) multiplication with constant A
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF4MULA(Q, A);
	input [3:0] A;
	output [3:0] Q;

	assign Q[0] = A[0];
	assign Q[1] = A[3];
	assign Q[2] = A[1];
	assign Q[3] = A[2];
endmodule

/*
* GF(4) multiplication with constant B
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF4MULB(Q, A);
	input [3:0] A;
	output [3:0] Q;

	assign Q[0] = A[0];
	assign Q[1] = A[2];
	assign Q[2] = A[3];
	assign Q[3] = A[1];	
endmodule


/*
* GF(3) square (x*x)
* @energy 0.39
* @area 91
* @delay 2
* @switch 1
*/
module GF3PWR2(Q, A);
	input [2:0] A;
	output [2:0] Q;

	wire w00;
	nor g00(w00, A[1], A[2]);
	not g10(Q[1], w00);
	assign Q[0] = A[0];
	assign Q[2] = 1'b0;
endmodule

/*
* GF(4) square (x*x)
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF4PWR2(Q, A);
	input [3:0] A;
	output [3:0] Q;

	assign Q[0] = A[0];
	assign Q[1] = A[1];
	assign Q[2] = A[3];
	assign Q[3] = A[2];
endmodule

/*
* GF(4) cube (x*x*x)
* @energy 0.44
* @area 109
* @delay 2
* @switch 1
*/
module GF4PWR3(Q, A);
	input [3:0] A;
	output [1:0] Q;

	wire w00;
	nor g00(w00, A[1], A[2], A[3]);
	not g10(Q[1], w00);
	assign Q[0] = A[0];
endmodule

/*
* Bus radix expansion
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF2TO3(Q, A);
	input [1:0] A;
	output [2:0] Q;

	assign Q[0] = A[0];
	assign Q[1] = A[1];
	assign Q[2] = 1'b0;
endmodule

/*
* Bus radix expansion
* @energy 0
* @area 0
* @delay 0
* @switch 0
*/
module GF2TO4(Q, A);
	input [1:0] A;
	output [3:0] Q;

	assign Q[0] = A[0];
	assign Q[1] = A[1];
	assign Q[2] = 1'b0;
	assign Q[3] = 1'b0;
endmodule

/*
* Quaternary to binary converter
* @energy 0.78
* @area 364
* @delay 2
* @switch 2
*/
module SPLITTER(Q0, Q1, A);
	input [3:0] A;
	output [1:0] Q0, Q1;
	
	wire w00, w01, w10, w11;
	nor g00(w00, A[0], A[2]);
	nor g01(w01, A[1], A[3]);
	nor g02(w10, A[0], A[1]);
	nor g03(w11, A[2], A[3]);
	not g10(Q0[0], w00);
	not g11(Q0[1], w01);
	not g12(Q1[0], w10);
	not g13(Q1[1], w11);
endmodule

/*
* Binary to quaternary converter
* @energy 0.35
* @area 364
* @delay 2
* @switch 1
*/
module MIXER(Q, A0, A1);
	input [1:0] A0, A1;
	output [3:0] Q;
	
	wire w00, w01, w10, w11;
	nand g00(w00, A0[0], A1[0]);
	nand g01(w01, A0[1], A1[0]);
	nand g02(w10, A0[0], A1[1]);
	nand g03(w11, A0[1], A1[1]);
	not g10(Q[0], w00);
	not g11(Q[1], w01);
	not g12(Q[2], w10);
	not g13(Q[3], w11);
endmodule
